@@ -2034,10 +2034,10 @@ pw_stream_connect(struct pw_stream *stream,
 	if ((str = getenv("PIPEWIRE_QUANTUM")) != NULL) {
 		struct spa_fraction q;
 		if (sscanf(str, "%u/%u", &q.num, &q.denom) == 2 && q.denom != 0) {
-			pw_properties_setf(stream->properties, PW_KEY_NODE_RATE,
+			pw_properties_setf(stream->properties, PW_KEY_NODE_FORCE_RATE,
 					"1/%u", q.denom);
-			pw_properties_setf(stream->properties, PW_KEY_NODE_LATENCY,
-					"%u/%u", q.num, q.denom);
+			pw_properties_setf(stream->properties, PW_KEY_NODE_FORCE_QUANTUM,
+					"%u", q.num);
 		}
 	}
 	if ((str = getenv("PIPEWIRE_LATENCY")) != NULL)
